from pyspark.sql.types import (
    IntegerType,
    StringType,
    StructField,
    StructType,
    BooleanType,
    LongType,
    DoubleType,
    FloatType,
    DataType,
)
from typing import List, Dict, Tuple, Callable, NamedTuple
from pyspark.sql import DataFrame
from streamstate_utils.structs import CassandraInputStruct, CassandraOutputStruct
import os


def get_folder_location(app_name: str, topic: str) -> str:
    return os.path.join(app_name, topic)


# get org name from ConfigMap
def get_cassandra_key_space_from_org_name(org_name: str) -> str:
    return org_name


def get_cassandra_table_name_from_app_name(app_name: str, version: str) -> str:
    return f"{app_name}_{version}"


ENV_NAMES = [
    "data_center",
    "cassandra_cluster_name",
    "port",
    "organization",
    "project",
    "org_bucket",
    "spark_namespace",
    "username",
    "password",
    # add checkpoint_location at some point
]


def _get_env_variables_from_config_map() -> dict:
    return {name: os.getenv(name, "") for name in ENV_NAMES}


def _convert_cluster_and_data_center_to_service_name(
    data_center: str, cassandra_cluster: str
) -> str:
    return f"{cassandra_cluster}-{data_center}-service"


def get_cassandra_inputs_from_config_map() -> CassandraInputStruct:
    env_var = _get_env_variables_from_config_map()
    return CassandraInputStruct(
        cassandra_ip=_convert_cluster_and_data_center_to_service_name(
            env_var["data_center"], env_var["cassandra_cluster_name"]
        ),
        cassandra_port=env_var["port"],
        cassandra_user=env_var["username"],
        cassandra_password=env_var["password"],
    )


# minorly inefficient
def get_organization_from_config_map() -> str:
    return _get_env_variables_from_config_map()["organization"]


def get_cassandra_outputs_from_config_map(
    app_name: str, version: str
) -> CassandraOutputStruct:
    env_var = _get_env_variables_from_config_map()
    return CassandraOutputStruct(
        cassandra_cluster=env_var["cassandra_cluster_name"],
        cassandra_key_space=get_cassandra_key_space_from_org_name(
            env_var["organization"]
        ),
        cassandra_table_name=get_cassandra_table_name_from_app_name(app_name, version),
    )


def get_output_from_config_map(app_name: str, version: str) -> CassandraOutputStruct:
    env_var = _get_env_variables_from_config_map()
    return CassandraOutputStruct(
        cassandra_cluster=env_var["cassandra_cluster_name"],
        cassandra_key_space=get_cassandra_key_space_from_org_name(
            env_var["organization"]
        ),
        cassandra_table_name=get_cassandra_table_name_from_app_name(app_name, version),
    )


def _convert_type(avro_type: str) -> DataType:
    avro_type_conversion = {
        "boolean": BooleanType(),
        "int": IntegerType(),
        "long": LongType(),
        "float": FloatType(),
        "double": DoubleType(),
        "string": StringType(),
    }
    return avro_type_conversion[avro_type]


def map_avro_to_spark_schema(fields: List[Dict[str, str]]) -> StructType:
    return StructType(
        [
            StructField(field["name"], _convert_type(field["type"]), True)
            for field in fields
        ]
    )



from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.streaming import StreamingQuery
import pyspark.sql.functions as F
from typing import List, Dict, Tuple, Callable
import sys
import shutil
from streamstate_utils.utils import map_avro_to_spark_schema, get_folder_location
import json
from streamstate_utils.structs import (
    OutputStruct,
    FileStruct,
    CassandraInputStruct,
    CassandraOutputStruct,
    KafkaStruct,
    InputStruct,
)


def kafka_wrapper(
    app_name: str,
    brokers: str,
    process: Callable[[List[DataFrame]], DataFrame],
    inputs: List[InputStruct],
    spark: SparkSession,
) -> DataFrame:
    dfs = [
        spark.readStream.format("kafka")
        .option("kafka.bootstrap.servers", brokers)
        .option("subscribe", input.topic)
        .load()
        .selectExpr("CAST(value AS STRING) as json")
        .select(
            F.from_json(
                F.col("json"), schema=map_avro_to_spark_schema(input.schema.fields)
            ).alias("data")
        )
        .select("data.*")
        for input in inputs
    ]
    return process(dfs)


def set_cassandra(
    cassandra: CassandraInputStruct,
    spark: SparkSession,
):
    spark.conf.set("spark.cassandra.connection.host", cassandra.cassandra_ip)
    spark.conf.set("spark.cassandra.connection.rpc.port", cassandra.cassandra_port)
    spark.conf.set("spark.cassandra.auth.username", cassandra.cassandra_user)
    spark.conf.set("spark.cassandra.auth.password", cassandra.cassandra_password)


def file_wrapper(
    app_name: str,
    max_file_age: str,
    process: Callable[[List[DataFrame]], DataFrame],
    inputs: List[InputStruct],
    spark: SparkSession,
) -> DataFrame:
    dfs = [
        spark.readStream.schema(map_avro_to_spark_schema(input.schema.fields))
        .option("maxFileAge", max_file_age)
        .json(get_folder_location(app_name, input.topic))
        for input in inputs
    ]
    return process(dfs)


def write_kafka(batch_df: DataFrame, kafka: KafkaStruct, output: OutputStruct):
    batch_df.write.format("kafka").option(
        "kafka.bootstrap.servers", kafka.brokers
    ).option("topic", output.output_name).save()


def write_parquet(batch_df: DataFrame, output_folder: str):
    batch_df.write.format("parquet").option("path", output_folder).save()


# make sure to call set_cassandra before this
def write_cassandra(batch_df: DataFrame, cassandra: CassandraOutputStruct):
    batch_df.write.format("org.apache.spark.sql.cassandra").option(
        "keyspace", cassandra.cassandra_key_space
    ).option("table", cassandra.cassandra_table_name).option(
        "cluster", cassandra.cassandra_cluster
    ).mode(
        "APPEND"
    ).save()


def write_console(
    result: DataFrame,
    checkpoint: str,
    mode: str,
):
    result.writeStream.format("console").outputMode("append").option(
        "truncate", "false"
    ).option("checkpointLocation", checkpoint).start().awaitTermination()


def write_wrapper(
    result: DataFrame,
    output: OutputStruct,
    write_fn: Callable[[DataFrame], None],
    # processing_time: str = "0",
):
    result.writeStream.outputMode(output.mode).option("truncate", "false").trigger(
        processingTime=output.processing_time
    ).option("checkpointLocation", output.checkpoint_location).foreachBatch(
        lambda df, id: write_fn(df)
    ).start().awaitTermination()



from typing import List, Dict, Tuple, Callable, Union
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.streaming import StreamingQuery
import shutil
import os
from streamstate_utils.generic_wrapper import file_wrapper
from streamstate_utils.structs import InputStruct
import json
import sys
import marshmallow_dataclass


def helper_for_file(
    app_name: str,
    max_file_age: str,
    process: Callable[[List[DataFrame]], DataFrame],
    inputs: List[InputStruct],
    spark: SparkSession,
    expecteds: List[dict],
):
    """
    This will be used for unit testing developer code
    Major inputs:

    process is the logic for manipulating streaming dataframes
    inputs is a list of "topic" names, example records from topic, and schema for topic
    expecteds is a list of expected output
    """
    file_dir = app_name
    try:
        shutil.rmtree(file_dir)
        print("folder exists, deleting")
    except:
        print("folder doesn't exist, creating")

    os.mkdir(file_dir)
    for input in inputs:
        os.mkdir(os.path.join(file_dir, input.topic))

    df = file_wrapper(
        app_name,
        max_file_age,
        process,
        inputs,
        spark,
    )
    file_name = "localfile.json"

    for input in inputs:
        file_path = os.path.join(file_dir, input.topic, file_name)
        with open(file_path, mode="w") as test_file:
            json.dump(input.sample, test_file)
    q = df.writeStream.format("memory").queryName(app_name).outputMode("append").start()
    try:

        assert q.isActive
        q.processAllAvailable()
        df = spark.sql(f"select * from {app_name}")
        result = df.collect()
        assert len(result) == len(expecteds)
        for row in result:
            dict_row = row.asDict()
            print(dict_row)
            assert dict_row in expecteds, f"{dict_row} is not in {expecteds}"
    finally:
        q.stop()
        shutil.rmtree(file_dir)



from dataclasses import dataclass, field
import marshmallow_dataclass
import marshmallow.validate
from marshmallow import Schema
from typing import ClassVar, Type, List, Dict

## Is output_name just the app_name?? that would make sense to me...
@dataclass
class OutputStruct:
    mode: str
    checkpoint_location: str
    output_name: str
    processing_time: str = "0"
    # primary_keys: List[str] = []  # not needed if dont persist to cassandra
    Schema: ClassVar[Type[Schema]] = Schema  # for mypy


@dataclass
class TableStruct:
    primary_keys: List[str]
    # organization: str
    output_schema: dict  # avro schema
    Schema: ClassVar[Type[Schema]] = Schema  # for mypy


@dataclass
class FileStruct:
    max_file_age: str
    Schema: ClassVar[Type[Schema]] = Schema  # for mypy


@dataclass
class SchemaStruct:
    fields: List[Dict[str, str]]
    Schema: ClassVar[Type[Schema]] = Schema  # for mypy


@dataclass
class InputStruct:
    topic: str
    schema: SchemaStruct
    sample: List[dict] = field(default_factory=list)  # not all need a sample
    Schema: ClassVar[Type[Schema]] = Schema  # for mypy


@dataclass
class CassandraInputStruct:
    cassandra_ip: str
    cassandra_port: str
    cassandra_password: str
    cassandra_user: str
    Schema: ClassVar[Type[Schema]] = Schema  # for mypy


@dataclass
class CassandraOutputStruct:
    cassandra_cluster: str
    cassandra_key_space: str
    cassandra_table_name: str
    Schema: ClassVar[Type[Schema]] = Schema  # for mypy


@dataclass
class KafkaStruct:
    brokers: str
    Schema: ClassVar[Type[Schema]] = Schema  # for mypy



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['streamstate_utils']

package_data = \
{'': ['*']}

install_requires = \
['marshmallow-dataclass>=8.4.1,<9.0.0']

setup_kwargs = {
    'name': 'streamstate-utils',
    'version': '0.2.1',
    'description': 'Utilities for cassandra and spark streaming specifically for streamstate',
    'long_description': None,
    'author': 'Daniel Stahl',
    'author_email': 'danstahl1138@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/StreamState/streamstate-utils',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
